from __future__ import annotations

import io
import struct
import sys
from typing import Literal, Union

import numpy as np

PCM = 0x0001
IEEE_FLOAT = 0x0003
EXTENSIBLE = 0xFFFE

AudioData = Union[np.memmap, np.ndarray]
Endian = Literal[">", "<"]  # Big Endian, Little Endian


def _read_fmt_chunk(
    fid: io.BufferedReader, en: Endian
) -> tuple[int, int, int, int, int]:
    size: int = struct.unpack(f"{en}I", fid.read(4))[0]

    if size < 16:
        raise ValueError("Binary structure of wave file is not compliant")

    res = struct.unpack(f"{en}HHIIHH", fid.read(16))
    bytes_read = 16

    format_tag, channels, sr, _, block_align, bit_depth = res
    # underscore is "bitrate"

    if format_tag == EXTENSIBLE and size >= 18:
        ext_chunk_size = struct.unpack(f"{en}H", fid.read(2))[0]
        bytes_read += 2
        if ext_chunk_size >= 22:
            extensible_chunk_data = fid.read(22)
            bytes_read += 22
            raw_guid = extensible_chunk_data[6:22]

            if en == ">":
                tail = b"\x00\x00\x00\x10\x80\x00\x00\xAA\x00\x38\x9B\x71"
            else:
                tail = b"\x00\x00\x10\x00\x80\x00\x00\xAA\x00\x38\x9B\x71"
            if raw_guid.endswith(tail):
                format_tag = struct.unpack(f"{en}I", raw_guid[:4])[0]
        else:
            raise ValueError("Binary structure of wave file is not compliant")

    if format_tag not in {PCM, IEEE_FLOAT}:
        raise ValueError(
            f"Encountered unknown format tag: {format_tag:#06x}, while reading fmt chunk."
        )

    # move file pointer to next chunk
    if size > bytes_read:
        fid.read(size - bytes_read)

    # fmt should always be 16, 18 or 40, but handle it just in case
    _handle_pad_byte(fid, size)

    return format_tag, channels, sr, block_align, bit_depth


def _read_data_chunk(
    fid: io.BufferedReader,
    format_tag: int,
    channels: int,
    bit_depth: int,
    en: Endian,
    block_align: int,
    data_size: int | None,
) -> AudioData:

    size: int = struct.unpack(f"{en}I", fid.read(4))[0]
    if data_size is not None:
        # size is only 32-bits here, so get real size from header.
        size = data_size

    bytes_per_sample = block_align // channels
    n_samples = size // bytes_per_sample

    if bytes_per_sample in (3, 5, 6, 7):
        raise ValueError(f"Unsupported bytes per sample: {bytes_per_sample}")

    if format_tag == PCM:
        if 1 <= bit_depth <= 8:
            dtype = "u1"  # WAVs of 8-bit integer or less are unsigned
        elif bit_depth <= 64:
            dtype = f"{en}i{bytes_per_sample}"
        else:
            raise ValueError(
                f"Unsupported bit depth: the WAV file has {bit_depth}-bit integer data."
            )
    elif format_tag == IEEE_FLOAT:
        if bit_depth in (32, 64):
            dtype = f"{en}f{bytes_per_sample}"
        else:
            raise ValueError(
                f"Unsupported bit depth: the WAV file has {bit_depth}-bit floating-point data."
            )
    else:
        raise ValueError(
            f"Unknown wave file format: {format_tag:#06x}. Supported formats: PCM, IEEE_FLOAT"
        )

    start = fid.tell()
    data = np.memmap(fid, dtype=dtype, mode="c", offset=start, shape=(n_samples,))
    fid.seek(start + size)

    _handle_pad_byte(fid, size)

    if channels > 1:
        try:
            _data = data.reshape(-1, channels)
        except ValueError:
            _data = data[:-1].reshape(-1, channels)
        return _data
    return data


def _skip_unknown_chunk(fid: io.BufferedReader, en: Endian) -> None:
    if data := fid.read(4):
        size = struct.unpack(f"{en}I", data)[0]
        fid.seek(size, 1)
        _handle_pad_byte(fid, size)


def _read_rf64_chunk(fid: io.BufferedReader) -> tuple[int, int, Endian]:

    # https://tech.ebu.ch/docs/tech/tech3306v1_0.pdf
    # https://www.itu.int/dms_pubrec/itu-r/rec/bs/R-REC-BS.2088-1-201910-I!!PDF-E.pdf

    heading = fid.read(12)
    if heading != b"\xff\xff\xff\xffWAVEds64":
        raise ValueError(f"Wrong heading: {heading!r}")

    chunk_size = fid.read(4)

    bw_size_low = fid.read(4)
    bw_size_high = fid.read(4)

    en: Endian = ">" if (bw_size_high > bw_size_low) else "<"

    data_size_low = fid.read(4)
    data_size_high = fid.read(4)

    # Combine bw_size and data_size to 64-bit ints

    def combine(a: bytes, b: bytes) -> int:
        return struct.unpack("<Q", a + b)[0]

    file_size = combine(bw_size_low, bw_size_high)
    data_size = combine(data_size_low, data_size_high)

    chunk_size = struct.unpack(f"{en}I", chunk_size)[0]
    fid.read(40 - chunk_size)  # type: ignore

    return data_size, file_size, en


def _read_riff_chunk(sig: bytes, fid: io.BufferedReader) -> tuple[None, int, Endian]:
    en: Endian = "<" if sig == b"RIFF" else ">"
    file_size: int = struct.unpack(f"{en}I", fid.read(4))[0] + 8

    form = fid.read(4)
    if form != b"WAVE":
        raise ValueError(f"Not a WAV file. RIFF form type is {form!r}.")

    return None, file_size, en


def _handle_pad_byte(fid: io.BufferedReader, size: int) -> None:
    if size % 2 == 1:
        fid.seek(1, 1)


def read(filename: str) -> tuple[int, AudioData]:
    fid = open(filename, "rb")

    try:
        file_sig = fid.read(4)
        if file_sig in (b"RIFF", b"RIFX"):
            data_size, file_size, en = _read_riff_chunk(file_sig, fid)
        elif file_sig == b"RF64":
            data_size, file_size, en = _read_rf64_chunk(fid)
        else:
            raise ValueError(f"File format {file_sig!r} not supported.")

        fmt_chunk_received = False
        data_chunk_received = False
        while fid.tell() < file_size:
            chunk_id = fid.read(4)

            if not chunk_id:
                if data_chunk_received:
                    break  # EOF but data successfully read
                raise ValueError("Unexpected end of file.")

            elif len(chunk_id) < 4 and not (fmt_chunk_received and data_chunk_received):
                raise ValueError(f"Incomplete chunk ID: {chunk_id!r}")

            if chunk_id == b"fmt ":
                fmt_chunk_received = True
                format_tag, channels, sr, block_align, bit_depth = _read_fmt_chunk(
                    fid, en
                )
            elif chunk_id == b"data":
                data_chunk_received = True
                if not fmt_chunk_received:
                    raise ValueError("No fmt chunk before data")

                data = _read_data_chunk(
                    fid,
                    format_tag,
                    channels,
                    bit_depth,
                    en,
                    block_align,
                    data_size,
                )
            else:
                _skip_unknown_chunk(fid, en)

    finally:
        fid.seek(0)

    return sr, data


def write(filename: str, sr: int, arr: AudioData) -> None:
    # Write empty samples to 'filename' with properties of given arr
    fid = open(filename, "wb")

    # Write RIFF WAV Header
    fid.write(b"RIFF\x00\x00\x00\x00WAVE")

    # Write RF64 WAV Header
    # fid.write(b"RF64\xff\xff\xff\xffWAVEds64")

    # #  - chunk_size
    # fid.write(b'\x1c\x00\x00\x00')  # Value based on bit-depth + '# of samples'

    # # - bw_size, Declaring Little Endian
    # fid.write(b'j@|\x00')
    # fid.write(b'\x00\x00\x00\x00')

    # Write 'fmt' Header
    dkind = arr.dtype.kind

    header_data = b"fmt "

    format_tag = IEEE_FLOAT if dkind == "f" else PCM
    channels = 1 if arr.ndim == 1 else arr.shape[1]
    bit_depth = arr.dtype.itemsize * 8
    bit_rate = sr * (bit_depth // 8) * channels
    block_align = channels * (bit_depth // 8)

    fmt_chunk_data = struct.pack(
        "<HHIIHH", format_tag, channels, sr, bit_rate, block_align, bit_depth
    )

    if not (dkind == "i" or dkind == "u"):
        # add cbSize field for non-PCM files
        fmt_chunk_data += b"\x00\x00"

    header_data += struct.pack("<I", len(fmt_chunk_data))
    header_data += fmt_chunk_data

    # fact chunk (non-PCM files)
    if not (dkind == "i" or dkind == "u"):
        header_data += b"fact"
        header_data += struct.pack("<II", 4, arr.shape[0])

    # check data size (needs to be immediately before the data chunk)
    if ((len(header_data) - 8) + (8 + arr.nbytes)) > 0xFFFFFFFF:
        raise ValueError("Data exceeds wave file size limit")

    fid.write(header_data)

    # Write Data Chunk
    fid.write(b"data")
    fid.write(struct.pack("<I", arr.nbytes))
    if arr.dtype.byteorder == ">" or (
        arr.dtype.byteorder == "=" and sys.byteorder == "big"
    ):
        arr = arr.byteswap()

    # Write the actual data
    fid.write(arr.ravel().view("b").data)

    # Write size info
    size = fid.tell()
    fid.seek(4)
    fid.write(struct.pack("<I", size - 8))

    fid.close()



data = {
    "Auto-Editor": {
        "_": """
Auto-Editor is an automatic video/audio creator and editor. By default, it will detect silence and create a new video with those sections cut out. By changing some of the options, you can export to a traditional editor like Premiere Pro and adjust the edits there, adjust the pacing of the cuts, and change the method of editing like using audio loudness and video motion to judge making cuts.

Run:
    auto-editor --help

To get the list of options.
""".strip(),
        "--set-speed-for-range": """
This option takes 3 arguments delimited with commas and they are as follows:
 - speed:
  - How fast to play the media (number)
Start:
  - The time when speed first gets applied (time)
End:
  - The time when speed stops being applied (time)

example:

--set-range-for-speed 2.5,400,800

will set the speed from 400 ticks to 800 ticks to 2.5x
If timebase is 30, 400 ticks to 800 means 13.33 to 26.66 seconds
""".strip(),
        "--edit-based-on": """
Editing Methods:
 - audio: General audio detection
 - motion: Motion detection specialized for real life noisy video
 - pixeldiff: Detect when a certain amount of pixels have changed between frames
 - random: Set silent/loud randomly based on a random or preset seed
 - none: Do not modify the media in anyway (Mark all sections as "loud")
 - all: Cut out everything out (Mark all sections as "silent")

Attribute Defaults:
 - audio
    - threshold: 4% (number)
    - stream: 0 (natural | "all")
 - motion
    - threshold: 2% (number)
    - stream: 0 (natural | "all")
    - blur: 9 (natural)
    - width: 400 (natural)
 - pixeldiff
    - threshold: 1 (natural)
    - stream: 0 (natural | "all")
 - random
    - threshold: 0.5 (number)
    - seed: RANDOMLY-GENERATED (int)

Operators:
 - and
   - usage: $METHOD and $METHOD
 - or
   - usage: $METHOD or $METHOD
 - xor
   - usage: $METHOD xor $METHOD
 - not
   - usage: not $METHOD
Examples:
  --edit audio
  --edit audio:stream=1
  --edit audio:threshold=4%
  --edit audio:threshold=0.03
  --edit motion
  --edit motion:threshold=2%,blur=3
  --edit audio:threshold=4% or motion:threshold=2%,blur=3
  --edit none
  --edit all
""".strip(),
        "--export": """
Instead of exporting a video, export as one of these options instead.

default       : Export as usual
premiere      : Export as an XML timeline file for Adobe Premiere Pro
final-cut-pro : Export as an XML timeline file for Final Cut Pro
shotcut       : Export as an XML timeline file for Shotcut
json          : Export as an auto-editor JSON timeline file
audio         : Export as a WAV audio file
clip-sequence : Export as multiple numbered media files
""".strip(),
        "--player": """
This option uses shell-like syntax to support using a specific player:

  auto-editor in.mp4 --player mpv

Args for the player program can be added as well:

  auto-editor in.mp4 --player 'mpv --keep-open'

Absolute or relative paths can also be used in the event the player's
executable can not be resolved:

  auto-editor in.mp4 --player '/path/to/mpv'
  auto-editor in.mp4 --player './my-relative-path/mpv'

If --player is not set, auto-editor will use the system default.
If --no-open is used, --player will always be ignored.

on MacOS, QuickTime can be used as the default player this way:

  auto-editor in.mp4 --player 'open -a "quicktime player"'
""".strip(),
        "--resolution": """
If not set, the first input's resolution is used.
If the input does not have a resolution (audio files) or there is not first input,
the value '1920x1080' will be used.
""".strip(),
        "--frame-rate": """
Set (timeline's) timebase and output frame rate.
It must be a string in the format frame_rate_num/frame_rate_den,
an integer, a floating point number, or a valid frame rate label.

The following labels are recognized:

  ntsc -> 30000/1001
  ntsc_film -> 24000/1001
  pal -> 25
  film -> 24
""".strip(),
        "--temp-dir": """
If not set, tempdir will be set with Python's tempfile module
The directory doesn't have to exist beforehand, however, the root path must be valid.
Beware that the temp directory can get quite big.
""".strip(),
        "--ffmpeg-location": "This takes precedence over `--my-ffmpeg`.",
        "--my-ffmpeg": "This is equivalent to `--ffmpeg-location ffmpeg`.",
        "--silent-threshold": """
Silent threshold is a percentage where 0% represents absolute silence and 100% represents the highest volume in the media file.
Setting the threshold to `0%` will cut only out areas where area is absolutely silence.
""".strip(),
        "--frame-margin": """
Margin is measured in frames if no units are specified. Seconds can be used. e.g. `0.3secs`
The starting and ending margins can be set separately with the use of a comma. e.g. `2sec,3sec` `7,10` `-1,6`
""".strip(),
        "--silent-speed": "99999 is the 'cut speed' and values over that or <=0 are considered 'cut speeds' as well",
        "--video-speed": "99999 is the 'cut speed' and values over that or <=0 are considered 'cut speeds' as well",
        "--min-clip-length": "Range: 0 to Infinity",
        "--min-cut-length": "Range: 0 to Infinity",
    },
    "info": {
        "_": "Retrieve information and properties about media files",
        "--include-vfr": """
A typical output will look like this:

- VFR:0.583394 (3204/2288) min: 41 max: 42 avg: 41

'0.583394' is the ratio of how many VFR frames are there.
'3204' is the number of VFR frames, '2288' is the number of non-VFR frames.
 Adding '3204' and '2288' will result in how many frames the video has in total.
""".strip(),
    },
    "levels": {"_": "Display loudness over time"},
    "subdump": {
        "_": "Dump text-based subtitles to stdout with formatting stripped out"
    },
    "grep": {"_": "Read and match text-based subtitle tracks"},
    "desc": {"_": "Display a media's description metadata"},
    "test": {"_": "Self-Hosted Unit and End-to-End tests"},
}



from __future__ import annotations

import difflib
import re
import sys
import textwrap
from dataclasses import dataclass
from shutil import get_terminal_size
from typing import Any, Callable, Iterator, Literal, TypeVar, Union

from auto_editor.utils.log import Log

T = TypeVar("T")
Nargs = Union[int, Literal["*"]]


@dataclass
class Required:
    names: tuple[str, ...]
    nargs: Nargs = "*"
    type: type = str
    choices: tuple[str, ...] | None = None
    metavar: str = "[file ...] [options]"


@dataclass
class Options:
    names: tuple[str, ...]
    nargs: Nargs = 1
    type: type = str
    flag: bool = False
    choices: tuple[str, ...] | None = None
    metavar: str | None = None
    help: str = ""


@dataclass
class OptionText:
    text: str


def indent(text: str, prefix: str) -> str:
    def predicate(line: str) -> str:
        return line.strip()

    def prefixed_lines() -> Iterator[str]:
        for line in text.splitlines(True):
            yield (prefix + line if predicate(line) else line)

    return "".join(prefixed_lines())


def out(text: str) -> None:
    width = get_terminal_size().columns - 3

    indent_regex = re.compile(r"^(\s+)")
    wrapped_lines = []

    for line in text.split("\n"):
        exist_indent = re.search(indent_regex, line)
        pre_indent = exist_indent.groups()[0] if exist_indent else ""

        wrapped_lines.append(
            textwrap.fill(line, width=width, subsequent_indent=pre_indent)
        )

    print("\n".join(wrapped_lines))


def print_program_help(reqs: list[Required], args: list[Options | OptionText]) -> None:
    text = f"Usage: {' '.join([req.metavar for req in reqs])}\n\nOptions:"

    width = get_terminal_size().columns - 3
    split = int(width * 0.44) + 3
    indent = "  "

    for i, arg in enumerate(args):
        if isinstance(arg, OptionText):
            if i == 0:
                text += f"\n  {arg.text}"
                indent = "    "
            else:
                text += f"\n\n  {arg.text}"
        else:
            text += "\n"
            line = f"{indent}{', '.join(reversed(arg.names))}"
            if arg.metavar is not None:
                line += f" {arg.metavar}"

            if arg.help == "":
                pass
            elif len(line) < split:
                line = textwrap.fill(
                    arg.help,
                    width=width,
                    initial_indent=f"{line}{' ' * (split - len(line))}",
                    subsequent_indent=split * " ",
                )
            else:
                line += "\n"
                line += textwrap.fill(
                    arg.help,
                    width=width,
                    initial_indent=split * " ",
                    subsequent_indent=split * " ",
                )

            text += line
    text += "\n\n"
    sys.stdout.write(text)


def to_underscore(name: str) -> str:
    """Convert new style options to old style.  e.g. --hello-world -> --hello_world"""
    return name[:2] + name[2:].replace("-", "_")


def to_key(op: Options | Required) -> str:
    """Convert option name to arg key.  e.g. --hello-world -> hello_world"""
    return op.names[0][:2].replace("-", "") + op.names[0][2:].replace("-", "_")


def print_option_help(program_name: str, ns_obj: T, option: Options) -> None:
    text = f"  {', '.join(option.names)} {'' if option.metavar is None else option.metavar}\n\n"

    bar_len = 11
    if option.flag:
        text += "    type: flag\n"
        bar_len = 15
    else:
        if option.nargs != 1:
            _add = f"    nargs: {option.nargs}\n"
            bar_len = len(_add)
            text += _add

        default: str | tuple | None = None
        try:
            default = getattr(ns_obj, to_key(option))
        except AttributeError:
            pass

        if default is not None:
            if isinstance(default, tuple):
                _add = f"    default: {','.join(map(str, default))}\n"
            else:
                _add = f"    default: {default}\n"
            bar_len = len(_add)
            text += _add

        if option.choices is not None:
            text += "    choices: " + ", ".join(option.choices) + "\n"

    text += f"    {'-' * (bar_len - 5)}\n\n"

    from auto_editor.help import data

    if option.names[0] in data[program_name]:
        text += indent(data[program_name][option.names[0]], "    ") + "\n"
    else:
        text += f"     {option.help}\n\n"

    out(text)


def get_option(name: str, options: list[Options]) -> Options | None:
    for option in options:
        if name in option.names or name in map(to_underscore, option.names):
            return option
    return None


def parse_value(option: Options | Required, val: str | None) -> Any:
    if val is None and option.nargs == 1:
        Log().error(f"{option.names[0]} needs argument.")

    try:
        value = option.type(val)
    except TypeError as e:
        Log().error(e)

    if option.choices is not None and value not in option.choices:
        my_choices = ", ".join(option.choices)

        Log().error(
            f"{value} is not a choice for {option.names[0]}\nchoices are:\n  {my_choices}"
        )

    return value


class ArgumentParser:
    def __init__(self, program_name: str) -> None:
        self.program_name = program_name
        self.requireds: list[Required] = []
        self.options: list[Options] = []
        self.args: list[Options | OptionText] = []

    def add_argument(self, *args: str, **kwargs) -> None:
        x = Options(args, **kwargs)
        self.options.append(x)
        self.args.append(x)

    def add_required(self, *args: str, **kwargs) -> None:
        self.requireds.append(Required(args, **kwargs))

    def add_text(self, text: str) -> None:
        self.args.append(OptionText(text))

    def parse_args(
        self,
        ns_obj: type[T],
        sys_args: list[str],
        macros: list[tuple[set[str], list[str]]] | None = None,
    ) -> T:
        if len(sys_args) == 0:
            from auto_editor.help import data

            out(data[self.program_name]["_"])
            sys.exit()

        if macros is not None:
            _macros = [(x[0].union(map(to_underscore, x[0])), x[1]) for x in macros]
            for old_options, new in _macros:
                for old_option in old_options:
                    if old_option in sys_args:
                        pos = sys_args.index(old_option)
                        sys_args[pos : pos + 1] = new
            del _macros
        del macros

        ns = ns_obj()
        option_names: list[str] = []

        program_name = self.program_name
        requireds = self.requireds
        options = self.options
        args = self.args

        builtin_help = Options(
            ("--help", "-h"),
            flag=True,
            help="Show info about this program or option then exit",
        )
        options.append(builtin_help)
        args.append(builtin_help)

        # Figure out command line options changed by user.
        used_options: list[Options] = []

        req_list: list[str] = []
        req_list_name = requireds[0].names[0]
        setting_req_list = requireds[0].nargs != 1

        oplist_name: str | None = None
        oplist_coerce: Callable[[str], str] = str

        i = 0
        while i < len(sys_args):
            arg = sys_args[i]
            option = get_option(arg, options)

            if option is None:
                if oplist_name is not None:
                    try:
                        val = oplist_coerce(arg)
                        ns.__setattr__(oplist_name, getattr(ns, oplist_name) + [val])
                    except (TypeError, ValueError) as e:
                        Log().error(e)
                elif requireds and not arg.startswith("--"):
                    if requireds[0].nargs == 1:
                        ns.__setattr__(req_list_name, parse_value(requireds[0], arg))
                        requireds.pop()
                    else:
                        req_list.append(arg)
                else:
                    label = "option" if arg.startswith("--") else "short"

                    # 'Did you mean' message might appear that options need a comma.
                    if arg.replace(",", "") in option_names:
                        Log().error(f"Option '{arg}' has an unnecessary comma.")

                    close_matches = difflib.get_close_matches(arg, option_names)
                    if close_matches:
                        Log().error(
                            f"Unknown {label}: {arg}\n\n    Did you mean:\n        "
                            + ", ".join(close_matches)
                        )
                    Log().error(f"Unknown {label}: {arg}")
            else:
                if option.nargs != "*":
                    if option in used_options:
                        Log().error(
                            f"Option {option.names[0]} may not be used more than once."
                        )
                    used_options.append(option)

                oplist_name = None
                oplist_coerce = option.type

                key = to_key(option)

                next_arg = None if i == len(sys_args) - 1 else sys_args[i + 1]
                if next_arg in ("-h", "--help"):
                    print_option_help(program_name, ns_obj, option)
                    sys.exit()

                if option.flag:
                    ns.__setattr__(key, True)
                elif option.nargs == 1:
                    ns.__setattr__(key, parse_value(option, next_arg))
                    i += 1
                else:
                    oplist_name = key

            i += 1

        if setting_req_list:
            ns.__setattr__(req_list_name, req_list)

        if getattr(ns, "help"):
            print_program_help(requireds, args)
            sys.exit()

        return ns



from __future__ import annotations

import sys
import tempfile
from dataclasses import dataclass, field
from fractions import Fraction

import numpy as np
from numpy.typing import NDArray

from auto_editor.analyze import (
    audio_levels,
    motion_levels,
    pixeldiff_levels,
    random_levels,
)
from auto_editor.ffwrapper import FFmpeg, FileInfo
from auto_editor.method import (
    audio_builder,
    motion_builder,
    pixeldiff_builder,
    random_builder,
)
from auto_editor.objects import _Vars, parse_dataclass
from auto_editor.output import Ensure
from auto_editor.utils.bar import Bar
from auto_editor.utils.log import Log
from auto_editor.utils.types import frame_rate
from auto_editor.vanparse import ArgumentParser


@dataclass
class Audio:
    stream: int


@dataclass
class Motion:
    stream: int
    blur: int
    width: int


@dataclass
class Pixeldiff:
    stream: int


@dataclass
class Random:
    seed: int


@dataclass
class LevelArgs:
    input: list[str] = field(default_factory=list)
    edit: str = "audio"
    timebase: Fraction | None = None
    ffmpeg_location: str | None = None
    my_ffmpeg: bool = False
    help: bool = False


def levels_options(parser: ArgumentParser) -> ArgumentParser:
    parser.add_required("input", nargs="*")
    parser.add_argument(
        "--edit",
        metavar="METHOD:[ATTRS?]",
        help="Select the kind of detection to analyze with attributes",
    )
    parser.add_argument(
        "--timebase",
        "-tb",
        metavar="NUM",
        type=frame_rate,
        help="Set custom timebase",
    )
    parser.add_argument("--ffmpeg-location", help="Point to your custom ffmpeg file")
    parser.add_argument(
        "--my-ffmpeg",
        flag=True,
        help="Use the ffmpeg on your PATH instead of the one packaged",
    )
    return parser


def print_floats(arr: NDArray[np.float_]) -> None:
    for a in arr:
        sys.stdout.write(f"{a:.20f}\n")


def print_ints(arr: NDArray[np.uint64]) -> None:
    for a in arr:
        sys.stdout.write(f"{a}\n")


def main(sys_args=sys.argv[1:]) -> None:
    parser = levels_options(ArgumentParser("levels"))
    args = parser.parse_args(LevelArgs, sys_args)

    ffmpeg = FFmpeg(args.ffmpeg_location, args.my_ffmpeg, False)

    bar = Bar("none")
    temp = tempfile.mkdtemp()
    log = Log(temp=temp)

    sources = {}
    for i, path in enumerate(args.input):
        sources[i] = FileInfo(path, i, ffmpeg, log)

    assert 0 in sources
    src = sources[0]

    tb = src.get_fps() if args.timebase is None else args.timebase
    ensure = Ensure(ffmpeg, src.get_samplerate(), temp, log)

    strict = True

    METHOD_ATTRS_SEP = ":"
    METHODS = ("audio", "motion", "pixeldiff", "random")

    if METHOD_ATTRS_SEP in args.edit:
        method, attrs = args.edit.split(METHOD_ATTRS_SEP)
    else:
        method, attrs = args.edit, ""

    if method not in METHODS:
        log.error(f"Method: {method} not supported")

    for src in sources.values():
        if method == "random":
            robj = parse_dataclass(attrs, (Random, random_builder[1:]), log)
            print_floats(random_levels(ensure, src, robj, tb, temp, log))

        if method == "audio":
            aobj = parse_dataclass(attrs, (Audio, audio_builder[1:]), log)
            print_floats(
                audio_levels(ensure, src, aobj.stream, tb, bar, strict, temp, log)
            )

        if method == "motion":
            if src.videos:
                _vars: _Vars = {"width": src.videos[0].width}
            else:
                _vars = {"width": 1}
            mobj = parse_dataclass(attrs, (Motion, motion_builder[1:]), log, _vars)
            print_floats(motion_levels(ensure, src, mobj, tb, bar, strict, temp, log))

        if method == "pixeldiff":
            pobj = parse_dataclass(attrs, (Pixeldiff, pixeldiff_builder[1:]), log)
            print_ints(pixeldiff_levels(ensure, src, pobj, tb, bar, strict, temp, log))

    log.cleanup()


if __name__ == "__main__":
    main()
